# -*- coding: utf-8 -*-
from unittest import TestCase
from unittest import mock

import eventisc
import eventisc.rabbit_listener


class TestRabbitListenerApp(TestCase):
    def tearDown(self):
        eventisc.default_app = None

    def _default_rabbit_config(self):
        return {
            "event_name_regex": ".*",
            "kind": "rabbit",
            "url": "http://rabbit:5672/end",
            "queue_kwargs": {
                "exchange": {
                    "exchange": "some-exchange",
                    "exchange_type": "direct"
                },
                "queue": {
                    "queue": "some-queue"
                }
            },
            "publish_kwargs": {
                "exchange": "{queue_kwargs['exchange']['exchange']}",
                "routing_key": "{queue_kwargs['queue']['queue']}"
            }
        }

    @mock.patch.object(eventisc.rabbit_listener, "pika")
    def test_rabbit_listener(self, pika_mock):
        app = eventisc.init_default_app(listeners=[self._default_rabbit_config()])

        channel_mock = app.listeners[0].local.channel

        app.trigger('client_created', {"id": 14, "topic": "client", "action": "created"})

        channel_mock.basic_publish.assert_called_once_with(
            exchange='some-exchange',
            routing_key='some-queue',
            body='{"id": 14, "topic": "client", "action": "created"}'
        )

    @mock.patch.object(eventisc.rabbit_listener, "pika")
    def test_pika_connection(self, pika_mock):
        app = eventisc.init_default_app(listeners=[self._default_rabbit_config()])

        channel_mock = app.listeners[0].local.channel

        channel_mock.queue_declare.assert_called_once_with(queue="some-queue")

        channel_mock.exchange_declare.assert_called_once_with(
            exchange="some-exchange",
            exchange_type="direct"
        )

        channel_mock.queue_bind.assert_called_once_with(
            exchange="some-exchange",
            queue="some-queue"
        )



# -*- coding: utf-8 -*-
import os
import tempfile
from unittest import TestCase
import eventisc

from unittest import mock


class MyTestListener(eventisc.Listener):
    kind = "test"

    def __init__(self, event_name, filter=None, foo=None):
        super().__init__(event_name, filter)
        self.foo = foo
        self.mock = mock.MagicMock()

    def _do_notify(self, event_name, event_data):
        return self.mock(event_name, event_data)


class TestEventApp(TestCase):
    def tearDown(self):
        eventisc.default_app = None

    def test_default_app(self):
        app = eventisc.get_current_app()
        assert app is not None

        with mock.patch.object(app, "trigger", wraps=app.trigger) as trigger_mock:
            app.trigger("myevent", {"foo": "bar"})
            trigger_mock.assert_called_once_with("myevent", {"foo": "bar"})
            trigger_mock.reset_mock()

            eventisc.trigger("myevent", {"foo": "bar"})
            trigger_mock.assert_called_once_with("myevent", {"foo": "bar"})

    def test_init_manual(self):
        my_listener = MyTestListener("testapp.myevent")
        app = eventisc.init_default_app("testapp.", [my_listener])
        assert app is not None
        assert len(app.listeners) == 1

        eventisc.trigger("fooevent", {"fii": "bar"})
        my_listener.mock.assert_not_called()

        eventisc.trigger("myevent", {"fii": "bar"})
        my_listener.mock.assert_called_once_with("testapp.myevent", {"fii": "bar"})

    @mock.patch.dict(os.environ, {"EVENTISC_NAME_PREFIX": "envname."})
    def test_init_from_dict(self):
        eventisc.init_default_app(listeners=[
            {"kind": "test", "event_name": "envname.myevent", "foo": 23},
            {"kind": "test", "event_name_regex": ".*_created", "foo": 45},
        ])

        app = eventisc.get_current_app()
        assert app.name_prefix == "envname."
        assert len(app.listeners) == 2

        assert isinstance(app.listeners[0], MyTestListener)
        assert app.listeners[0].foo == 23
        assert app.listeners[1].foo == 45

        count = eventisc.trigger("fooevent", {"fii": "bar"})
        assert count == 0

        count = eventisc.trigger("myevent", {"fii": "bar"})
        assert count == 1
        app.listeners[0].mock.assert_called_once_with("envname.myevent", {"fii": "bar"})
        app.listeners[1].mock.assert_not_called()
        app.listeners[0].mock.reset_mock()

        count = eventisc.trigger("bar_created", {"bar_name": "Moe's"})
        assert count == 1
        app.listeners[0].mock.assert_not_called()
        app.listeners[1].mock.assert_called_once_with("envname.bar_created", {"bar_name": "Moe's"})

    def _make_temp_file(self, content, suffix=".yaml"):
        ret = tempfile.NamedTemporaryFile(suffix=suffix, mode="wt")
        ret.write(content)
        ret.flush()
        return ret

    def test_init_from_json(self):
        config_file = self._make_temp_file("""{"name_prefix": "jsonapp.",
            "listeners": [{"kind": "test", "event_name": "jsonapp.myevent", "foo": 33}]
        }""", suffix=".json")

        with mock.patch.dict(os.environ, {"EVENTISC_CONFIG": config_file.name}):
            app = eventisc.get_current_app()
            assert app.name_prefix == "jsonapp."
            assert len(app.listeners) == 1
            assert app.listeners[0].foo == 33

    def test_init_from_yaml(self):
        config_file = self._make_temp_file("""name_prefix: yamlapp.
listeners:
 - kind: test
   event_name: jsonapp.myevent
   foo: 666
""")

        with mock.patch.dict(os.environ, {"EVENTISC_CONFIG": config_file.name}):
            app = eventisc.get_current_app()
            assert app.name_prefix == "yamlapp."
            assert len(app.listeners) == 1
            assert app.listeners[0].foo == 666

    def test_init_from_yaml_and_nameprefix(self):
        config_file = self._make_temp_file("""listeners:
 - kind: test
   event_name: jsonapp.myevent
   foo: 666
""")

        with mock.patch.dict(os.environ, {"EVENTISC_CONFIG": config_file.name,
                                          "EVENTISC_NAME_PREFIX": "envapp."}):
            app = eventisc.get_current_app()
            assert app.name_prefix == "envapp."
            assert len(app.listeners) == 1
            assert app.listeners[0].foo == 666

    def test_filter_expression(self):
        eventisc.init_default_app(listeners=[
            {"kind": "test", "event_name": "myevent", "foo": 23,
             "filter": {"kind": "expr", "expr": "event_data['fii'] > 10"}},
        ])

        app = eventisc.get_current_app()
        assert len(app.listeners) == 1

        assert isinstance(app.listeners[0], MyTestListener)
        assert app.listeners[0].foo == 23

        count = eventisc.trigger("myevent", {"fii": 34})
        assert count == 1
        app.listeners[0].mock.assert_called_once_with("myevent", {"fii": 34})
        app.listeners[0].mock.reset_mock()

        count = eventisc.trigger("myevent", {"fii": 9})
        assert count == 0
        app.listeners[0].mock.assert_not_called()

    @mock.patch.dict(os.environ, {"EVENTISC_DRYRUN": "Y"})
    def test_dry_run(self):
        eventisc.init_default_app(listeners=[
            {"kind": "test", "event_name": "myevent", "foo": 23},
        ])

        app = eventisc.get_current_app()
        assert len(app.listeners) == 1

        assert isinstance(app.listeners[0], MyTestListener)
        assert app.listeners[0].foo == 23

        with mock.patch.object(eventisc.logger, "info", wraps=eventisc.logger.info) as info_log:
            count = eventisc.trigger("myevent", {"fii": 34})
            assert count == 1
            app.listeners[0].mock.assert_not_called()
            info_log.assert_called_once_with('EVENTISC_DRYRUN=Y _do_notify(%s, %s)',
                                             'myevent', {'fii': 34})



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from . import Listener
import pika
import json
import logging
import threading
from functools import wraps

logger = logging.getLogger(__name__)


def reconnection_handler(f):
    @wraps(f)
    def decorator(self, *args, **kwargs):
        try:
            ret = f(self, *args, **kwargs)
        except pika.exceptions.AMQPConnectionError:
            logger.info("Error connecting RabbitMQ, deleting channel...")
            del self.local.channel
            ret = f(self, *args, **kwargs)
        return ret
    return decorator


class RabbitListener(Listener):
    kind = "rabbit"

    def __init__(self, event_name, url, queue_kwargs=None, publish_kwargs=None, filter=None):
        super(RabbitListener, self).__init__(event_name, filter)
        self.url = url
        self.queue_kwargs = queue_kwargs
        self.publish_kwargs = publish_kwargs

        self.local = threading.local()
        self.lock = threading.Lock()
        self.connection = None

        self.declare_queue_and_exchange()

    @reconnection_handler
    def declare_queue_and_exchange(self):
        queue = self.queue_kwargs.get("queue", {})
        self.get_channel().queue_declare(**queue)
        exchange = self.queue_kwargs.get("exchange", {})
        if exchange:
            self.get_channel().exchange_declare(**exchange)
            self.get_channel().queue_bind(exchange=exchange.get("exchange"), queue=queue.get("queue"))

    def get_channel(self):
        self._connect()

        if hasattr(self.local, 'channel'):
            return self.local.channel

        try:
            self.local.channel = self.connection.channel()
        except pika.exceptions.AMQPConnectionError:
            logger.info("Error connecting RabbitMQ, getting new connection...")
            self._connect(force_reconnect=True)
            self.local.channel = self.connection.channel()

        return self.local.channel

    def _connect(self, force_reconnect=False):
        # thread safe connection aquire
        with self.lock:
            if force_reconnect:
                self.connection = None

            if self.connection is not None:
                return

            self.connection = pika.BlockingConnection(pika.URLParameters(self.url))

        if not hasattr(self.local, "channel"):
            self.local.channel = self.connection.channel()

    @reconnection_handler
    def _do_notify(self, event_name, event_data):
        if self.publish_kwargs:
            data = dict((k, self.format(v, event_name, event_data, queue_kwargs=self.queue_kwargs))
                        for (k, v) in self.publish_kwargs.items())
        else:
            data = None
        data["body"] = json.dumps(event_data)
        self.get_channel().basic_publish(**data)
